-----------------------------------------------------------------------------
--
-- Tag				:= { Statement } @
-- Statement		:= literal | "[" RValue "]"
-- RValue			:= Relation
-- Relation			:= Expression [ ("==" | "~=" | ">" | "<" | "<=" | ">=") Expression ]
-- Expression		:= Term [ ("+" | "-") Term ]
-- Factor			:= Chain [ ("*" | "/") Chain ]
-- Chain			:= Primary { ":" Primary }
-- Primary			:= ConstantNumber | ConstantString | Identifier ["(" ParameterList ")"] | ( RValue )
-- ParameterList	:= RValue {, RValue }
--
-----------------------------------------------------------------------------

TagCompiler = { }

-----------------------------------------------------------------------------
--
-- CreateTag passed TagCompilerData structure that optionally has:
--  tagName.data
--  tagName.events
--  tagName.overrideEvents
--  tagName.guaranteeResult
--  tagName.noGuaranteeChange
--  tagName.numeric				<-- only set to true if the value is also guaranteed
-- And must have one of:
--  tagName.tableData
--  tagName.string
--  tagName.method
--  tagName.compileString
--
-----------------------------------------------------------------------------

local tagCompilerData
local parseTag
local parseTagOffset
local tagMethods = { }
local overrideEvents = { }
local numeric = false
local guaranteeResult

local function GetByte()
	while parseTag:byte(parseTagOffset) == 32 do 
		parseTagOffset = parseTagOffset + 1
	end 
	return parseTag:byte(parseTagOffset) or 0
end

local function AdvanceByte()
	parseTagOffset = parseTagOffset + 1
end

-----------------------------------------------------------------------------

local function TagErrorStringMethod() return "!!Tag Error!!" end
local function TagError(x) tagCompilerData.TagError(x) end
local function TagDebug(x) tagCompilerData.TagDebug(x) end

local function EmptyTable(table) for k in pairs(table) do table[k] = nil end end
local function MergeData(targetTable, sourceData)
	if not sourceData then return end
	if type(sourceData) == "string" then
		targetTable[sourceData] = true
	elseif type(sourceData) == "table" then
		for _,v in pairs(sourceData) do
			targetTable[v] = true
		end
	end
end

-----------------------------------------------------------------------------

function TagCompiler:CoerceToNumber(text)
	if not text or text == "" then return "0" end
	if numeric then return text end
	return ("(%s or 0)"):format(text)
end

function TagCompiler:ParseParameterList(dataTable)
	if GetByte() == 40 then -- byte == '('
		AdvanceByte()
		local byte = GetByte()
		if byte == 41 then 
			AdvanceByte()
			return nil
		end

		local parameterListStringList = { self:ParseRValue(dataTable) }

		while parseTagOffset <= #parseTag and GetByte() == 44 do -- byte == ','
			AdvanceByte()
			table.insert(parameterListStringList, self:ParseRValue(dataTable))
		end

		if GetByte() == 41 then -- byte == ')'
			AdvanceByte()
		else
			dataTable.error = true
			TagError("Missing end of parameter list \")\"")
		end
		
		return parameterListStringList
	else
		return nil
	end
end

function TagCompiler:ParsePrimary(dataTable, functionString)
	local byte = GetByte()
	if byte >= 48 and byte <= 59 then -- byte >= '0' && byte <= '9'
		local startOffset = parseTagOffset
		local endOffset = startOffset
		local decimalPointFound = false
		while true do
			endOffset = endOffset + 1
			byte = parseTag:byte(endOffset) or 0
			if byte == 46 then -- '.'
				if decimalPointFound then
					break
				else
					decimalPointFound = true
				end
			elseif byte < 48 or byte > 59 then
				break
			end
		end
		parseTagOffset = endOffset
		if functionString then
			numeric = false
			return ("(%s and %s)"):format(functionString, parseTag:sub(startOffset, endOffset-1))
		else
			numeric = true
			guaranteeResult = true
			return parseTag:sub(startOffset, endOffset-1)
		end
	elseif byte == 34 then	-- byte == '\"'
		local startOffset = parseTagOffset+1
		local endOffset = startOffset + 1
		numeric = false
		while true do
			if endOffset > #parseTag then
				TagError("Missing closing quote")
				dataTable.error = true
				return ""
			end
			byte = parseTag:byte(endOffset) or 0
			if byte == 34 then -- byte == '\"'
				break
			elseif byte == 92 then	-- byte == '\\'
				endOffset = endOffset + 2
			else
				endOffset = endOffset + 1
			end
		end
		parseTagOffset = endOffset+1
		if functionString then
			return ("%s and %q"):format(functionString, parseTag:sub(startOffset, endOffset-1))
		else
			guaranteeResult = true
			return ("%q"):format(parseTag:sub(startOffset, endOffset-1))
		end
		
	elseif (byte >= 65 and byte <= 90) or (byte >= 97 and byte <= 122) then	-- byte >= 'A' && byte <= 'Z' || byte >= 'a' && byte <= 'z'
		local startOffset = parseTagOffset
		local endOffset = startOffset
		while (byte >= 65 and byte <= 90) or (byte >= 97 and byte <= 122) do
			endOffset = endOffset + 1
			byte = parseTag:byte(endOffset) or 0
		end
		parseTagOffset = endOffset
		local identifier = parseTag:sub(startOffset, endOffset-1):lower()

		local tagData = tagCompilerData[identifier]
		if not tagData then
			TagError("Invalid tag: "..identifier)
			dataTable.error = true
			return ""
		end

		MergeData(dataTable.data, tagData.data)
		MergeData(dataTable.events, tagData.events)
		MergeData(overrideEvents, tagData.overrideEvents)

		if tagData.string then
			local tag, offset = parseTag, parseTagOffset
			parseTag, parseTagOffset = tagData.string, 1
			local returnValue = self:ParsePrimary(dataTable, functionString)
			parseTag, parseTagOffset = tag, offset
			return returnValue
		end

		numeric = tagData.numeric
		if not tagData.noGuaranteeChange then
			guaranteeResult = tagData.guaranteeResult
		end

		if tagData.tableData then
			if functionString then
				return functionString.." and data."..tagData.tableData
			else
				return "data."..tagData.tableData
			end
		elseif tagData.method then
			tagMethods[identifier] = true	-- this marks the method as required, and it's inserted as an upvalue
			local parameterStringList = self:ParseParameterList(dataTable)
			if parameterStringList then
				return ("tagMethod%s(data, %s, %s)"):format(identifier, functionString or "true", table.concat(parameterStringList, ", "))
			else
				return ("tagMethod%s(data, %s)"):format(identifier, functionString or "true")
			end
		elseif tagData.compileString then
			local parameterStringList = self:ParseParameterList(dataTable)
			return tagData.compileString(dataTable, functionString, parameterStringList and unpack(parameterStringList) or nil)
		end
	elseif (byte == 40) then -- '('
		AdvanceByte()
		local result = self:ParseRValue(dataTable)
		local byte = GetByte()
		if byte == 41 then
			AdvanceByte()
			return ("(%s)"):format(result)
		else
			TagError("Missing \")\"")
			dataTable.error = true
			return ""
		end
	end
end

function TagCompiler:ParseChain(dataTable)
	local functionString = self:ParsePrimary(dataTable)
	while GetByte() == 58 do	-- while tag[offset] == ':'
		AdvanceByte()
		functionString = self:ParsePrimary(dataTable, functionString)
	end
	return functionString
end

function TagCompiler:ParseFactor(dataTable)
	local result = self:ParseChain(dataTable)
	local byte = GetByte()
	while byte == 42 or byte == 47 do	-- '*' or '/'
		AdvanceByte()
		result = self:CoerceToNumber(result)
		result = ("%s %s %s"):format(result, string.char(byte), self:CoerceToNumber(self:ParseFactor(dataTable)))
		numeric = true
		byte = GetByte()
	end
	return result
end

function TagCompiler:ParseExpression(dataTable)
	local result = self:ParseFactor(dataTable)
	local byte = GetByte()
	while byte == 43 or byte == 45 do	-- '+' or '-'
		AdvanceByte()
		result = self:CoerceToNumber(result)
		result = ("%s %s %s"):format(result, string.char(byte), self:CoerceToNumber(self:ParseFactor(dataTable)))
		numeric = true
		byte = GetByte()
	end
	return result
end

function TagCompiler:ParseRelation(dataTable)
	local result = self:ParseExpression(dataTable)

	local byte = GetByte()
	local byte2 = parseTag:byte(parseTagOffset+1)
	if byte == 61 then	-- '='
		AdvanceByte()
		if byte2 == 61 then
			AdvanceByte()
			return result.."=="..self:ParseExpression(dataTable)
		else
			TagError("Unexpected \"=\"")
			dataTable.error = true
			return ""
		end
	elseif byte == 126 then -- '~'
		AdvanceByte()
		if byte2 == 61 then
			AdvanceByte()
			return result.."~="..self:ParseExpression(dataTable)
		else
			TagError("Unexpected \"=\"")
			dataTable.error = true
			return ""
		end
	elseif byte == 60 then -- '<'
		AdvanceByte()
		result = self:CoerceToNumber(result)
		if byte2 == 61 then
			AdvanceByte()
			return ("%s <= %s"):format(result, self:CoerceToNumber(self:ParseExpression(dataTable)))
		else
			return ("%s < %s"):format(result, self:CoerceToNumber(self:ParseExpression(dataTable)))
		end
	elseif byte == 62 then -- '>'
		AdvanceByte()
		result = self:CoerceToNumber(result)
		if byte2 == 61 then
			AdvanceByte()
			return ("%s >= %s"):format(result, self:CoerceToNumber(self:ParseExpression(dataTable)))
		else
			return ("%s > %s"):format(result, self:CoerceToNumber(self:ParseExpression(dataTable)))
		end
	else
		return result
	end
end

function TagCompiler:ParseRValue(dataTable)
	local gr = guaranteeResult
	guaranteeResult = false
	local returnValue = self:ParseRelation(dataTable)
	guaranteeResult = guaranteeResult and gr
	return returnValue
end

-- Initialise the compile cache with an empty tag, something which would otherwise need some extra special-cased code
local compileCache =
{
	[""] = { data = { }, events = { }, tag = "", method = function() return "" end },
}

local headerTable = { }
local functionStringList = { }
function TagCompiler:CreateTag(tag, tagData)
	parseTag = tag or ""
	tagCompilerData = tagData

	if compileCache[parseTag] then
		return compileCache[parseTag]
	end

	parseTagOffset = 1
	local dataTable = { data = { }, events = { }, tag = parseTag }

	table.insert(functionStringList, "") -- it's replaced with all the locals required later
	table.insert(functionStringList, "return function(data)\nreturn ")

	while parseTagOffset <= #parseTag and not dataTable.error do
		if parseTagOffset > 1 then
			table.insert(functionStringList, "..\n ")
		end
		if parseTag:byte(parseTagOffset) == 91 then	-- if tag[offset] == '['
			AdvanceByte()
			guaranteeResult = true
			table.insert(functionStringList, ([[(%s%s)]]):format(self:ParseRValue(dataTable) or "", guaranteeResult and "" or " or \"\""))
			if GetByte() == 93 then -- if tag[offset] == ']'
				AdvanceByte()
			elseif not dataTable.error then
				TagError("Tag not closed - missing \"]\"")
				dataTable.error = true
			end
		else
			-- literal section
			local startOffset = parseTagOffset
			local endOffset = startOffset+1
			while endOffset <= #parseTag and parseTag:byte(endOffset) ~= 91 do
				endOffset = endOffset + 1
			end
			table.insert(functionStringList, ("%q"):format(parseTag:sub(startOffset, endOffset-1)))
			parseTagOffset = endOffset
		end
	end

	for k in pairs(overrideEvents) do
		dataTable.events[k] = nil
		overrideEvents[k] = nil
	end

	if dataTable.error then
		dataTable.method = TagErrorStringMethod
		EmptyTable(dataTable.data)
		EmptyTable(dataTable.events)
	else
		self.tagCompilerData = tagData
		for k in pairs(tagMethods) do
			table.insert(headerTable, ("local tagMethod%s = TagCompiler.tagCompilerData[\"%s\"].method\n"):format(k, k))
		end
		functionStringList[1] = table.concat(headerTable)
		EmptyTable(headerTable)
		table.insert(functionStringList, "\nend")
		local functionString = table.concat(functionStringList)

--[[
		TagDebug(string.format("Tag: %s", parseTag))
		for event in pairs(dataTable.events) do
			TagDebug(string.format(" - event: %s", event))
		end
		for data in pairs(dataTable.data) do
			TagDebug(string.format(" - data: %s", data))
		end
		TagDebug("--- Compiled method ---\n"..functionString)
--]]--
		dataTable.method = loadstring(functionString)()
		compileCache[parseTag] = dataTable
	end

	EmptyTable(tagMethods)
	EmptyTable(functionStringList)

	return dataTable
end

-----------------------------------------------------------------------------
